@@ -35,10 +35,7 @@
 #include <glib.h>
 #include <gio/gunixinputstream.h>
 #include <glib-unix.h>
-
-#ifdef ENABLE_X11_SUPPORT
 #include <X11/Xauth.h>
-#endif
 
 #define DISPLAY_FILENO (STDERR_FILENO + 1)
 #define BUS_ADDRESS_FILENO (DISPLAY_FILENO + 1)
@@ -117,7 +114,7 @@ prepare_auth_file (void)
         GError   *error = NULL;
         gboolean  prepared = FALSE;
         Xauth     auth_entry = { 0 };
-        char      localhost[_POSIX_HOST_NAME_MAX + 1] = "";
+        char      localhost[HOST_NAME_MAX + 1] = "";
 
         g_debug ("Preparing auth file for X server");
 
@@ -127,7 +124,7 @@ prepare_auth_file (void)
                 return NULL;
         }
 
-        if (gethostname (localhost, _POSIX_HOST_NAME_MAX) < 0) {
+        if (gethostname (localhost, HOST_NAME_MAX) < 0) {
                 strncpy (localhost, "localhost", sizeof (localhost) - 1);
         }
 
@@ -263,16 +260,16 @@ spawn_x_server (State        *state,
          * by default anymore. In older versions we need to pass
          * -nolisten tcp to disable listening on tcp sockets.
          */
-        if (!allow_remote_connections) {
-                g_ptr_array_add (arguments, "-nolisten");
-                g_ptr_array_add (arguments, "tcp");
-        }
-
-#ifdef HAVE_XSERVER_WITH_LISTEN
+#ifdef HAVE_XSERVER_THAT_DEFAULTS_TO_LOCAL_ONLY
         if (allow_remote_connections) {
                 g_ptr_array_add (arguments, "-listen");
                 g_ptr_array_add (arguments, "tcp");
         }
+#else
+        if (!allow_remote_connections) {
+                g_ptr_array_add (arguments, "-nolisten");
+                g_ptr_array_add (arguments, "tcp");
+        }
 #endif
 
         g_ptr_array_add (arguments, "-background");
